/***
StackPivotChecker: A BTF based StackPivoting Exploit Detection System
!!! This code is for education purpose and just demo code on Windows 7 32bit OS!!!
Copyright (c) 2012-2016, Intel Corporation
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; Version 2.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
Contact information:
xiaoning.li@intel.com
*/ 


#include <ntddk.h>           
#include <string.h>

#include "btf_driver.h"

#define NT_DEVICE_NAME      L"\\Device\\btf_device"
#define DOS_DEVICE_NAME     L"\\DosDevices\\btf_device"

#pragma pack(1)
typedef struct {
	UINT16  limit;
	UINT32  base;
} IA32_GDTR, IA32_IDTR;

typedef struct {
  UINT32  OffsetLow:16;   // Offset bits 15..0
  UINT32  Selector:16;    // Selector
  UINT32  Reserved_0:8;   // Reserved
  UINT32  GateType:8;     // Gate Type.  See #defines above
  UINT32  OffsetHigh:16;  // Offset bits 31..16
} IA32_IDT_GATE_DESCRIPTOR;

// use nonpagd memory
#pragma data_seg("NONPAGE")

//device open status
UINT32 openstatus = 0;

//current CR3
UINT32 currentCR3 = 0;

//we only support one CR3 to use BTF device at same time so far. you can extend it as array
UINT32 CallerCR3 = 0;
// new idt status
UINT32 vidt_installed = 0;

//idt data
static IA32_IDTR idt_desc = {0};
//default OS #DB handler
UINT32 OS_DB_Handler = 0;

#pragma data_seg()

//read idtr
void ia32_read_idtr(IA32_IDTR *p_descriptor)
{
	__asm
	{
		mov   edx, p_descriptor
		sidt  [edx]
	}
}

//write idtr
void ia32_write_idtr(IA32_IDTR *p_descriptor)
{
	__asm
	{
		
		push edx;
		mov   edx, p_descriptor;
		cli;
		lidt  fword ptr [edx];
		sti;
		pop edx;
		
	}
}

// stackpivot handler as new #DB handler
__declspec(naked) void stackpivotchecker(void)
{
	
	__asm{

		push eax
		push ebx
		push ecx
		push edx

		; only handle ring 3 code from CS
		mov eax,[esp+14h] 
		test eax, 1h
		jz OS_HANDLER_R0

		; only handle target cr3
		lea eax, currentCR3
		mov eax,[eax]
		mov ebx,cr3
		cmp eax,ebx
		jnz OS_HANDLER_R0

		; enable BTF/LBR to force BTF every time
		mov ecx, 1d9h
		rdmsr
		or eax,2h  
		wrmsr
		
		;get stack range from Thread Information Block, may cause PF			
		mov eax,fs:[18h]
		;stacktop
		mov ecx,[eax+4]
		;stackbottom
		mov edx,[eax+8]
					
		;get ring3 ESP
		mov ebx,[esp+1ch]
					
		cmp ebx,ecx
		ja  STHCHK
		cmp edx,ebx
		ja  STHCHK
		jmp DUMP_HANDLER
		
STHCHK:					
		;force NX bit to stop the code execution and cause pf crash in ring 3
		mov eax, [esp+10h]
		shr     eax, 0Ch
		shl     eax, 3
		and     eax, 7FFFF8h
		sub     eax, 40000000h
		mov     dword ptr [eax+4], 80000000h
		mov     eax, cr3
		mov     cr3, eax
					
DUMP_HANDLER:
		pop edx
		pop ecx
		pop ebx
		pop eax
		iretd

OS_HANDLER_R0:

		pop edx
		pop ecx
		pop ebx
		pop eax
		jmp  [OS_DB_Handler]
	}

}

//get cr3
 void GetCR3(UINT32* cr3value)
{
	__asm{
		push eax
		push edx
		mov eax,cr3
		;mov edx,[esp+0xc]
		mov edx, cr3value
		mov [edx],eax
		pop edx
		pop eax
	}

}

// print IDT vector handlers
void print_idt_vector(IA32_IDT_GATE_DESCRIPTOR* Idt, int vectorid)
{
    
	DbgPrint("BTF_Driver:\tVector[%x] Type 0x%x Selector:Address 0x%x:0x%x ", vectorid,Idt[vectorid].GateType, Idt[vectorid].Selector, (Idt[vectorid].OffsetHigh << 16) | (Idt[vectorid].OffsetLow));

}

// install vector handler
UINT32 InstallVectorHandler(IA32_IDT_GATE_DESCRIPTOR* Idt, UINT32 vectorid, UINT32 HandlerAddr)
{
	UINT32 PreviousAddr = (Idt[vectorid].OffsetHigh <<16) + Idt[vectorid].OffsetLow;
    Idt[vectorid].OffsetLow  = HandlerAddr & 0xFFFF;
    Idt[vectorid].OffsetHigh = HandlerAddr >> 16;
	return PreviousAddr;
}

//install new #DB handler for all cores
void install_new_singstep_handler(void)
{
	 
	
	KAFFINITY ActiveProcessors, CurrentAffinity;
	int cpuindex = 0;
	 
	DbgPrint("BTF_Driver:\tStart to hook IDT\n");

	ActiveProcessors = KeQueryActiveProcessors();
	DbgPrint("BTF_Driver:\tActiveProcessors = %x \n",ActiveProcessors);

	   for (CurrentAffinity = 1; ActiveProcessors; CurrentAffinity <<= 1) 
	   {
		    
		 if (ActiveProcessors & CurrentAffinity)
		 {
		    ActiveProcessors &= ~CurrentAffinity;
		    KeSetSystemAffinityThread(CurrentAffinity);

			ia32_read_idtr(&idt_desc);
			DbgPrint("BTF_Driver:\tCurrent IDT limit = %X, base = %p \n",idt_desc.limit,idt_desc.base);

			print_idt_vector((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1);

			OS_DB_Handler = InstallVectorHandler((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1,(UINT32)stackpivotchecker);
			
			print_idt_vector((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1);

		    KeRevertToUserAffinityThread();
		    cpuindex++;
		   
		}
		  
	   }
		
		return;

}

//uninstall single step handler and recover OS default one
void uninstall_new_singstep_handler(void)
{
	KAFFINITY ActiveProcessors, CurrentAffinity;
	int cpuindex = 0;
	 
	DbgPrint("BTF_Driver:\tStart to unhook IDT\n");

	ActiveProcessors = KeQueryActiveProcessors();
	DbgPrint("BTF_Driver:\tActiveProcessors = %x \n",ActiveProcessors);

	   for (CurrentAffinity = 1; ActiveProcessors; CurrentAffinity <<= 1) 
	   {
		    
		 if (ActiveProcessors & CurrentAffinity)
		 {
		    ActiveProcessors &= ~CurrentAffinity;
		    KeSetSystemAffinityThread(CurrentAffinity);

			ia32_read_idtr(&idt_desc);
			DbgPrint("BTF_Driver:\tCurrent IDT limit = %X, base = %p \n",idt_desc.limit,idt_desc.base);

			print_idt_vector((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1);

			if(OS_DB_Handler)
			{
				InstallVectorHandler((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1,(UINT32)OS_DB_Handler);
			}
			
			print_idt_vector((IA32_IDT_GATE_DESCRIPTOR*)idt_desc.base,1);

		
		   KeRevertToUserAffinityThread();
		   cpuindex++;
		   
		}
		  
	   }

	OS_DB_Handler = 0;
	
	return;
}

	
DRIVER_INITIALIZE DriverEntry;


NTSTATUS
DriverEntry(
    __in PDRIVER_OBJECT   DriverObject,
    __in PUNICODE_STRING      RegistryPath
    )

{
    NTSTATUS        ntStatus;
    UNICODE_STRING  DeviceName;   
    UNICODE_STRING  DosDeviceName; 
    PDEVICE_OBJECT  deviceObject = NULL;

    UNREFERENCED_PARAMETER(RegistryPath);

    RtlInitUnicodeString( &DeviceName, NT_DEVICE_NAME );

    ntStatus = IoCreateDevice(
        DriverObject,                   
        0,                              
        &DeviceName,               
        FILE_DEVICE_UNKNOWN,            
        FILE_DEVICE_SECURE_OPEN,     
        FALSE,                       
        &deviceObject );             

    if ( !NT_SUCCESS( ntStatus ) )
    {
        DbgPrint("BTF_Driver Couldn't create the device object\n");
        return ntStatus;
    }

    //
    // Initialize the driver object with this driver's entry points.
    //

    DriverObject->MajorFunction[IRP_MJ_CREATE] = BTF_Create;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = BTF_Close;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = BTF_DeviceControl;
    DriverObject->DriverUnload = BTF_UnloadDriver;

    //
    // Initialize a Unicode String containing the Win32 name
    // for our device.
    //

    RtlInitUnicodeString( &DosDeviceName, DOS_DEVICE_NAME );

    //
    // Create a symbolic link between our device name  and the Win32 name
    //

    ntStatus = IoCreateSymbolicLink(
                        &DosDeviceName, &DeviceName );

    if ( !NT_SUCCESS( ntStatus ) )
    {
        //
        // Delete everything that this routine has allocated.
        //
        DbgPrint("BTF_Driver Couldn't create symbolic link\n");
        IoDeleteDevice( deviceObject );
    }


	MmLockPagableCodeSection((PVOID )((UINT32)stackpivotchecker));
	
    return ntStatus;
}


NTSTATUS
BTF_Create(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
{
    UNREFERENCED_PARAMETER(DeviceObject);

	DbgPrint("BTF_Driver:\tCalled BTF_Create\n");
	

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}

NTSTATUS
BTF_Close(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
{
    UNREFERENCED_PARAMETER(DeviceObject);

	DbgPrint("BTF_Driver:\tCalled BTF_Close\n");
	
	if(openstatus)
	{

		GetCR3(&CallerCR3);
		if(CallerCR3==currentCR3)
		{

			if(vidt_installed)
			{
				uninstall_new_singstep_handler();
				vidt_installed = 0;
			}

			if(currentCR3!=0)
			{
				currentCR3 = 0;
			}
			openstatus = 0;

			
		}
	}
		
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}


VOID
BTF_UnloadDriver(
    __in PDRIVER_OBJECT DriverObject
    )
{
    PDEVICE_OBJECT deviceObject = DriverObject->DeviceObject;
    UNICODE_STRING uniWin32NameString;

	MmUnlockPagableImageSection((PVOID )((UINT32)stackpivotchecker));
	
    RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );

    IoDeleteSymbolicLink( &uniWin32NameString );

    if ( deviceObject != NULL )
    {
        IoDeleteDevice( deviceObject );
    }



}

NTSTATUS
BTF_DeviceControl(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
{
    PIO_STACK_LOCATION  irpSp;// Pointer to current stack location
    NTSTATUS            status = STATUS_SUCCESS;// Assume success
    
    UNREFERENCED_PARAMETER(DeviceObject);

    irpSp = IoGetCurrentIrpStackLocation( Irp );
    
    //
    // Determine which I/O control code was specified.
    //

    switch ( irpSp->Parameters.DeviceIoControl.IoControlCode )
    {
    case IOCTL_BTF_BTFTRACE_OPEN:

        DbgPrint("BTF_Driver:\tCalled IOCTL_BTF_BTFTRACE_OPEN\n");
	
		if(!openstatus)
		{
			if(currentCR3==0)
			{

				GetCR3(&currentCR3);

				DbgPrint("currentCR3 %x\n",currentCR3);
			
				openstatus = 1;
				
				if(!vidt_installed)
				{
					install_new_singstep_handler();
					
					vidt_installed = 1;
				}
			
			}
		}
        
      break;

    case IOCTL_BTF_BTFTRACE_CLOSE:

        DbgPrint("BTF_Driver:\tCalled IOCTL_BTF_BTFTRACE_CLOSE\n");

        
		if(openstatus)
		{

			GetCR3(&CallerCR3);
			if(CallerCR3==currentCR3)
			{

				if(vidt_installed)
				{
					uninstall_new_singstep_handler();
					
					vidt_installed = 0;

				}

				if(currentCR3!=0)
				{
					currentCR3 = 0;
				}
				openstatus = 0;

				
			}
		}
		
        
      break;

    default:

        //
        // The specified I/O control code is unrecognized by this driver.
        //

        status = STATUS_INVALID_DEVICE_REQUEST;
        DbgPrint("BTF_Driver:\tERROR: unrecognized IOCTL %x\n",
            irpSp->Parameters.DeviceIoControl.IoControlCode);
        break;
    }


    Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = 0;
    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return status;
}
